#include "io.h"
#include "log.h"
#include "shmem.h"
#include "filed.h"
#include "main.h"
#include "externs.h"
#include "flags.h"
#include "socket.h"
#include "session.h"
#include "url.h"
#include "peer.h"
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

void process_link(struct connection *c, struct parse *p) {
	char *arg1;
	char *arg2;
	struct session *s;
	char *file;
	char *url;
	
	arg1 = get_token_s(p,1); /* session */
	arg2 = get_token_s(p,2); /* file */

	if(!arg1 || !arg2) {
		c_send(c,"ERR ULINK \"S or A requires two arguments\"");
		return;
	}

	/* now we must either perform the link ourselves or proxy it */
	/* fortunately we already know which we want to do */

	/* lets start by getting a session object */
		
	/* if we are NOT locally handling this, we want to pass the request
	   verbetum to the machine that is (since it might have a different
	   underlying filesystem than we do */

	/* if we aren't locally handling this, we don't get to send the return
	   data ourselves.. at least not yet. */

	if(filed->best_peer != 0) 
		if(peer_proxy_to(filed->best_peer,"SX %d \"%s\" \"%s\"",c->id,arg1,arg2)) 
			return;

	if(filed->peer_flags & PFLAG_DOWN) {
		c_send(c,"ERR \"Unable to service your request\"");
		return;
	}

	file = generate_file(arg2);
	url = generate_url(arg1,arg2);

	if(!url || !file) {
		c_send(c,"ERR \"Unable to create link for file - may be bad media type\"");

		return;
	}

	/* attempt to stat the file, return an error if stat fails */
	if(!file_check(file)) {
		c_send(c,"ERR \"File [%s] does not exist\"",file);
		return;
	}

	s = get_session(arg1);

	if(!s) {
		c_send(c,"ERR ULINK \"Unable to create session object\"");
		return;
	}

	s->last_access = time(NULL);

	shmem_file_add(file,url,s->session_seq);
	url = finalize_url(url);
	if(!url) {
		c_send(c,"ERR \"Unable to create final url for file - may be bad media type\"");
		return;
	}

	c_send(c,"RD %s",url);
	return;	
}

char *finalize_url(char *url) {
	static char ret[256];
	char *ch;
	struct parse *p;

	if(!url) return(NULL);
	p = get_media_type(url);
	
	if(!(ch = get_token_s(p,2))) 
		return(NULL);
	strcpy(ret,ch);
	strcat(ret,url);
	return(ret);
}


char *generate_file(char *file) {
	/* look up this media type in the conf file and return the base path for it */
	static char ret[256];
	struct parse *p;
	char *c;

	p = get_media_type(file);
	c = get_token_s(p,3);
	if(!c) return(NULL);
	strcpy(ret,c);
	strcat(ret,file);
	return(ret);
}

char *generate_url(char *session, char *file) {
	/* generate a URL for requested file _file_ */
	/* this URL should include a locally generated sequence # */
	/* as well as padded session ID and file path */
	/* all '/' or '\' should be replaced by '-' in the file */

	static char ret[256];
	char sid[40];
	char *ch;
	
	
	if(!session || !file) {
		elog("generate_url called with invalid args");
		die();
	}

	
	/* look up the media type for this file and return the base URL to it */
	
	strcpy(ret,"");
	sprintf(sid,"%d",get_request_id());
	strcat(ret,pad(sid,11,'0'));
	strcat(ret,pad(session,20,'_'));
	strcat(ret,pad(file,15,'_'));
	ch = url_encode(ret);
	return(ch);
}

char *url_encode(char *url) {
	int a,b,c,d;
	static char buf[256];
	char data[256];

	dlog(DEBUG_CODE,"url_encode() called with [%s]",url);
	
	if(!url) return(NULL);
	strcpy(data,url);

	if(sys_flags & FLAG_NOCRYPT) {
		a = 0;
		b = 0;
	} else {
		a = (int)(10.0*rand()/(RAND_MAX+1.0));
                b = (int)(10.0*rand()/(RAND_MAX+1.0));
        }

	for(c=0;c<strlen(url);c++) {
		if(data[c] == '.') break;
		switch(data[c]) {
			case '/':
				data[c] = '-';
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				d = data[c] - '0';
					if(c % 2) 
					d+=a;
					else
					d+=b;
				if(d>9) d-=10;
				data[c] = d + '0';
				break;
			default:
				break; /* do nothing */
			}
		}
		sprintf(buf,"%d%d%s",a,b,data);
		dlog(DEBUG_CODE,"url_encode() returned [%s]",buf);
		return(buf);
}
	
struct parse *get_media_type(char *mt) {
	struct parse_hdr *p = filed->cf;
	struct parse *r;
	char *mta;
	char *mtb;

	if(!mt) return(NULL);

	if(!(mta = strrchr(mt,'.')))
		return(NULL);


	while((p = get_entry(p,"media-type"))) {
		r = p->data;
		mtb = get_token_s(r,1);
		if(mtb && !strcasecmp(mtb,mta))
			return(r);
	p = p->next;
	}

	return(NULL);
}

char *pad(char *input, int num, char padchr) {
	static char ret[256];
	int pad;
	char pc[2];
	int len = strlen(input);
	pad = num - len;
	
	dlog(DEBUG_CODE,"pad called: %s %d %d %c",input,num,pad,padchr);
	
	if(pad <= 0) {
		strncpy(ret,input,256);
		return(ret);
	}

	strcpy(ret,"");
	sprintf(pc,"%c",padchr);

	while(pad--)
		strcat(ret,pc);
	strcat(ret,input);

	dlog(DEBUG_CODE,"pad returned: %s",ret);
	return(ret);
}

int get_request_id(void) {
	int ret;

	ret = ++filed->request_id;
	if(ret > 65534) ret = 0;
	return(ret);
}


int file_check(char *file) {
	struct stat sb;
	if(stat(file,&sb)) {
		return(0);
	}

	if(!S_ISREG(sb.st_mode) && !S_ISLNK(sb.st_mode))
		return(0);

	return(1);
}

void url_process_sx(struct connection *c, struct parse *p) {
	char *arg0;
	char *arg1;
	char *arg2;
	struct session *s;
	char *file;
	char *url;
	
	arg0 = get_token_s(p,1); /* return_session */
	arg1 = get_token_s(p,2); /* session */
	arg2 = get_token_s(p,3); /* file */

	if(!arg0 || !arg1 || !arg2) {
		c_send(c,"ERR ULINK \"SX requires three arguments\"");
		dlog(DEBUG_PEER,"url_process_sx(): peer %s (%d) tried to call SX with [%s] [%s] [%s]",c->name,c->id,arg0,arg1,arg2);
		return;
	}

	file = generate_file(arg2);
	url = generate_url(arg1,arg2);

	if(!url || !file) {
		c_send(c,"%s \"ERR Unable to create link for file - may be bad media type\"",arg0);

		return;
	}

	/* attempt to stat the file, return an error if stat fails */
	if(!file_check(file)) {
		c_send(c,"%s \"ERR File [%s] does not exist\"",arg0,file);
		return;
	}

	s = get_session(arg1);

	if(!s) {
		c_send(c,"%s \"ERR ULINK Unable to create session object\"",arg0);
		return;
	}

	s->last_access = time(NULL);

	shmem_file_add(file,url,s->session_seq);
	url = finalize_url(url);
	if(!url) {
		c_send(c,"%s \"ERR Unable to create final url for file - may be bad media type\"",arg0);
		return;
	}

	c_send(c,"TX %s \"RD %s\"",arg0,url);
	return;	
}
